import sys
import argparse
import logging
import getpass
import os
from . import samTobed
from . import pyWriter
from . import henipipe

POLL_TIME = 5
LOG_PREFIX = '[HENIPIPE]: '

# Set up a basic logger
LOGGER = logging.getLogger('something')
myFormatter = logging.Formatter('%(asctime)s: %(message)s')
handler = logging.StreamHandler()
handler.setFormatter(myFormatter)
LOGGER.addHandler(handler)
LOGGER.setLevel(logging.DEBUG)
myFormatter._fmt = "[HENIPIPE]: " + myFormatter._fmt


def run_henipipe(args=None):
    if args is None:
        args = sys.argv[1:]
    parser = argparse.ArgumentParser('A wrapper for running henipipe')
    parser.add_argument('job', type=str, choices=['MAKERUNSHEET', 'ALIGN', 'SCALE', 'MERGE', 'SEACR', 'MACS2', 'AUC', 'GENOMESFILE', 'FASTQC'], help='a required string denoting segment of pipeline to run.  1) "MAKERUNSHEET" - to parse a folder of fastqs; 2) "ALIGN" - to perform alignment using bowtie and output bed files; 3) "SCALE" - to normalize data to reference (spike in); 4) "MERGE" - to merge bedgraphs 5) "SEACR" - to perform SEACR; 6) "MACS" - to perform MACS2; 7) "AUC" - to calculate AUC between normalized bedgraph using a peak file; 8) "GENOMESFILE" - print location of genomes.json file; 9) "FASTQC" - run fastqc on cluster')
    parser.add_argument('--sample_flag', '-sf', type=str, default="", help='FOR MAKERUNSHEET only string to identify samples of interest in a fastq folder')
    parser.add_argument('--fastq_folder', '-fq', type=str, help='For MAKERUNSHEET only: Pathname of fastq folder (files must be organized in folders named by sample)')
    parser.add_argument('--genome_key', '-gk', type=str, default="default", help='For MAKERUNSHEET only: abbreviation to use "installed" genomes in the runsheet (See README.md for more details')
    parser.add_argument('--filter_high', '-fh', type=int, default=None, help='For ALIGN only: upper limit of fragment size to exclude, defaults is no upper limit.  OPTIONAL')
    parser.add_argument('--filter_low', '-fl', type=int, default=None, help='For ALIGN only: lower limit of fragment size to exclude, defaults is no lower limit.  OPTIONAL')
    parser.add_argument('--output', '-o', type=str, default=".", help='For MAKERUNSHEET only: Pathname to write runsheet.csv file (folder must exist already!!), Defaults to current directory')
    parser.add_argument('--runsheet', '-r', type=str, help='tab-delim file with sample fields as defined in the script. - REQUIRED for all jobs except MAKERUNSHEET')
    parser.add_argument('--log_prefix', '-l', type=str, default='henipipe.log', help='Prefix specifying log files for henipipe output from henipipe calls. OPTIONAL')
    parser.add_argument('--select', '-s', type=str, default=None, help='To only run the selected row in the runsheet, OPTIONAL')
    parser.add_argument('--debug', '-d', action='store_true', help='To print commands (For testing flow). OPTIONAL')
    parser.add_argument('--bowtie_flags', '-b', type=str, default='--end-to-end --very-sensitive --no-mixed --no-discordant -q --phred33 -I 10 -X 700', help='For ALIGN: bowtie flags, OPTIONAL')
    parser.add_argument('--cluster', '-c', type=str, default='SLURM', choices=['PBS', 'SLURM', 'local'], help='Cluster software.  OPTIONAL Currently supported: PBS, SLURM and local')
    parser.add_argument('--threads', '-t', type=str, default=None, help='number of threads')
    parser.add_argument('--gb_ram', '-gb', type=str, default=None, help='gigabytes of RAM per thread')
    parser.add_argument('--install', '-i', type=str, default=None, help='FOR GENOMESFILE: location of file to install as a new genomes.json file, existing genomes.json will be erased')
    parser.add_argument('--norm_method', '-n', type=str, default='coverage', choices=['coverage', 'read_count', 'spike_in'], help='For ALIGN and SCALE: Normalization method, by "read_count", "coverage", or "spike_in".  If method is "spike_in", HeniPipe will align to the spike_in reference genome provided in runsheet. OPTIONAL')
    parser.add_argument('--user', '-u', type=str, default=None, help='user for submitting jobs - defaults to username.  OPTIONAL')
    parser.add_argument('--SEACR_norm', '-Sn', type=str, default='non', choices=['non', 'norm'], help='For SEACR: Normalization method; default is "non"-normalized, select "norm" to normalize using SEACR. OPTIONAL')
    parser.add_argument('--SEACR_stringency', '-Ss', type=str, default='stringent', choices=['stringent', 'relaxed'], help='FOR SEACR: Default will run as "stringent", other option is "relaxed". OPTIONAL')
    parser.add_argument('--keep_files', '-k', action ='store_true', default=False, help='FOR ALIGN: use this flag to turn off piping (Will generate all files).')
    parser.add_argument('--verbose', '-v', default=False, action='store_true', help='Run with some additional ouput - not much though... OPTIONAL')
    """
    call = 'henipipe MAKERUNSHEET -fq ../fastq -sf mini -gk heni_hg38 -o .'
    call = 'henipipe MACS2 -r ./runsheet.csv -d -mk -s 1:10'
    call = 'henipipe GENOMESFILE'
    call = 'henipipe MAKERUNSHEET -fq ../fastq'
    call = 'henipipe MAKERUNSHEET -fq ../fastq'
    call = 'henipipe ALIGN -r runsheet.csv -d'
    args = parser.parse_args(call.split(" ")[1:])
    """
    args = parser.parse_args()

    if args.job=="GENOMESFILE":
        _ROOT = os.path.abspath(os.path.dirname(__file__))
        if args.install is None:
            GENOMES_JSON = os.path.join(_ROOT, 'data', 'genomes.json')
            print(GENOMES_JSON)
        if args.install is not None:
            from shutil import copyfile
            args.install = os.path.abspath(args.install)
            copyfile(args.install, os.path.join(_ROOT, 'data', 'genomes.json'))
        exit()
    #log

    #deal with user
    if args.user is None:
        args.user = getpass.getuser()

    #deal with paths
    if args.job=="MAKERUNSHEET":
        if os.path.isabs(args.fastq_folder) is False:
            if args.fastq_folder == ".":
                args.fastq_folder = os.getcwd()
            else :
                args.fastq_folder = os.path.abspath(args.fastq_folder)
        if os.path.exists(args.fastq_folder) is False:
            raise ValueError('Path: '+args.fastq_folder+' not found')
        if os.path.isabs(args.output) is False:
            if args.output == ".":
                args.output = os.getcwd()
            else :
                args.output = os.path.abspath(args.output)
        if os.path.exists(args.output) is False:
            raise ValueError('Path: '+args.output+' not found')
    if args.job != "MAKERUNSHEET":
        if os.path.exists(args.runsheet) is False:
            raise ValueError('Path: '+args.runsheet+' not found')
        args.output = os.path.abspath(args.output)


    if args.job=="MAKERUNSHEET":
        LOGGER.info("Parsing fastq folder - "+args.fastq_folder+" ...")
        LOGGER.info("Writing runsheet to - "+os.path.join(args.output, 'runsheet.csv')+" ...")
        LOGGER.info("Using genome_key - "+args.genome_key+" ...")
        henipipe.make_runsheet(folder=args.fastq_folder, output=args.output, sample_flag = args.sample_flag, genome_key = args.genome_key, no_pipe=args.keep_files)
        exit()

    #parse and chech runsheet
    args.runsheet = os.path.abspath(args.runsheet)

    """
    parsed_runsheet = list(parse_runsheet(args.runsheet))
    check_runsheet(args, parsed_runsheet, verbose=args.verbose)
    """
    parsed_runsheet = list(henipipe.parse_runsheet(args.runsheet))

    henipipe.check_runsheet(args, parsed_runsheet, verbose=args.verbose)

    #deal with sample selection
    if args.select is not None:
        parsed_runsheet = [parsed_runsheet[i-1] for i in list(henipipe.parse_range_list(args.select))]

    if args.debug == False:
        LOGGER.info("Logging to %s... examine this file if samples fail." % args.log_prefix)

    if args.job=="FASTQC":
        LOGGER.info("Running fastqc on all fastqs in runsheet")
        Fastqcjob = henipipe.Fastqc(runsheet_data = parsed_runsheet, threads = args.threads, gb_ram = args.gb_ram, debug=args.debug, cluster=args.cluster, log=args.log_prefix, user=args.user)
        Fastqcjob.run_job()
        exit()

    if args.job=="ALIGN":
        #deal with filtering
        LOGGER.info("Aligning reads...")




        #Alignjob = Align(runsheet_data = parsed_runsheet, threads = args.threads, gb_ram = args.gb_ram, debug=args.debug, no_pipe=args.keep_files, cluster=args.cluster, bowtie_flags=args.bowtie_flags, log=args.log_prefix, user=args.user, norm_method=args.norm_method, filter = [args.filter_low, args.filter_high])


        Alignjob = henipipe.Align(runsheet_data = parsed_runsheet, threads = args.threads, gb_ram = args.gb_ram, debug=args.debug, no_pipe=args.keep_files, cluster=args.cluster, bowtie_flags=args.bowtie_flags, log=args.log_prefix, user=args.user, norm_method=args.norm_method, filter = [args.filter_low, args.filter_high])
        LOGGER.info("Submitting alignment jobs... Debug mode is %s" % args.debug)
        Alignjob.run_job()
        exit()

    if args.job=="SCALE":
        LOGGER.info("Calculating %s", args.norm_method)
        Scalejob = henipipe.Scale(runsheet_data = parsed_runsheet, threads = args.threads, gb_ram = args.gb_ram, debug=args.debug, cluster=args.cluster, log=args.log_prefix, norm_method=args.norm_method, user=args.user)
        LOGGER.info("Submitting bedgraph jobs... Debug mode is %s" % args.debug)
        Scalejob.run_job()
        exit()

    if args.job=="MERGE":
        Mergejob = henipipe.Merge(runsheet_data = parsed_runsheet, threads = args.threads, gb_ram = args.gb_ram, debug=args.debug, cluster=args.cluster, log=args.log_prefix, norm_method=args.norm_method, user=args.user, out=args.output)
        #Mergejob = Merge(runsheet_data = parsed_runsheet, debug=args.debug, cluster=args.cluster, log=args.log_prefix, norm_method=args.norm_method, user=args.user)
        LOGGER.info("Submitting merge-bedgraph jobs... Debug mode is %s" % args.debug)
        Mergejob.run_job()
        exit()

    if args.job=="SEACR":
        LOGGER.info("Running SEACR using settings: SEACR_norm = %s, SEACR_stringency = %s" % (args.SEACR_norm, args.SEACR_stringency))
        SEACRjob = henipipe.SEACR(runsheet_data = parsed_runsheet, threads = args.threads, gb_ram = args.gb_ram, debug=args.debug, cluster=args.cluster, norm=args.SEACR_norm, stringency=args.SEACR_stringency, user=args.user, log=args.log_prefix)
        SEACRjob.run_job()
        exit()

    if args.job=="MACS2":
        LOGGER.info("Running MACS2")
        MACS2job = henipipe.MACS2(runsheet_data = parsed_runsheet, threads = args.threads, gb_ram = args.gb_ram, debug=args.debug, cluster=args.cluster, user=args.user, log=args.log_prefix, out=args.output)
        MACS2job.run_job()
        exit()

    if args.job=="AUC":
        LOGGER.info("Running AUC")
        AUCjob = henipipe.AUC(runsheet_data = parsed_runsheet, threads = args.threads, gb_ram = args.gb_ram, debug=args.debug, no_pipe=args.keep_files, cluster=args.cluster, user=args.user, log=args.log_prefix, out=args.output, norm=args.SEACR_norm, stringency=args.SEACR_stringency)
        AUCjob.run_job()
        exit()


if __name__ == "__main__":
    run_henipipe()

"""
[parsed_runsheet[i-1] for i in list(parse_range_list("1:4,11,12"))]

"""






#!/usr/bin/python
# PBS/SLURM cluster job submission in Python

# A wrapper that performs the necessary pipeline for generation of CNR/CNT data
# Written by Scott Furlan with code inspiration from Andrew Hill's cellwrapper; 
# uses a custom python script samTobed.py which takes code from a fantastic 
# sam reader "simplesam" - https://github.com/mdshw5/simplesam

# runsheet columns
# 1 - 'sample' name of the sample REQUIRED
# 2 - 'fasta' location of the Bowtie2 indexed fasta file REQUIRED
# 3 - 'spikein_fasta' location of the Bowtie2 indexed fasta file for spike_in normalization OPTIONAL
# 4 - 'fastq1' a tab seperated string of filenames denoting location of all R1 files for a sample REQUIRED
# 5 - 'fastq2' a tab seperated string of filenames denoting location of all R2 files for a sample REQUIRED
# 6 - 'bed_out' name of the location for the aligned and sorted bam file REQUIRED
# 7 - 'spikein_bed_out' name of the location for the aligned and sorted bam file OPTIONAL
# 8 - 'genome_sizes' REQUIRED
# 9 - 'bedgraph' file name of normalized bedgraph REQUIRED
# 10 - 'SEACR_key' sample key corresponding to sample groups to be run against an IgG (or other) contol.  all samples to be run against a control are given the same name and the control is labeleled with the an additional flag +'_CONTROL' (i.e. 4JS_CONTROL) OPTIONAL
# 11 - 'SEACR_out' file name of SEACR output OPTIONAL

import os
from subprocess import Popen, PIPE
import argparse
import time
import sys
import csv
import re
import string
import random
from itertools import chain, compress
import json
from datetime import datetime


#_ROOT = os.getcwd()
_ROOT = os.path.abspath(os.path.dirname(__file__))
#_ROOT = '/Users/sfurla/Box Sync/PI_FurlanS/computation/develop/henipipe/henipipe'
GENOMES_JSON = os.path.join(_ROOT, 'data', 'genomes.json')
SEACR_SCRIPT = os.path.join(_ROOT, 'scripts', 'SEACR_1.3.sh')
ENVIRONS_JSON = os.path.join(_ROOT, 'data', 'environs.json')



class SampleFactory:
    def __init__(self, *args, **kwargs):
        self.environs = Environs(cluster = kwargs.get('cluster'), user = kwargs.get('user'), log = kwargs.get('log'), threads = kwargs.get('threads'), gb_ram = kwargs.get('gb_ram'))

        #remove later
        self.user = kwargs.get('user')
        self.cluster = kwargs.get('cluster')
        self.threads = kwargs.get('threads')
        self.gb_ram = kwargs.get('gb_ram')
        self.cluster = kwargs.get('cluster')

        self.runsheet_data = kwargs.get('runsheet_data')
        self.debug = kwargs.get('debug')
        # self.log_name = kwargs.get('log')
    def __call__():
        pass

    def id_generator(self, size=10, chars=string.ascii_uppercase + string.digits):
        return ''.join(random.choice(chars) for _ in range(size))

    def run_job(self):
        popen_command = self.environs.popen_command
        with open('out.log') as out_log, open('err.log') as err_log:
            for script in self.bash_scripts:
                if self.debug==False:
                    # Open a pipe to the command.
                    proc = Popen(popen_command, shell=True, stdin=PIPE, stdout=PIPE, stderr=PIPE, close_fds=True)
                    if (sys.version_info > (3, 0)):
                        proc.stdin.write(script.encode('utf-8'))
                        out, err = proc.communicate()
                    else:
                        proc.stdin.write(script)
                        out, err = proc.communicate()
                # Print your job and the system response to the screen as it's submitted
                print(script)
                if self.debug==False:
                    print(err)
                    print(out)
                    out_log.write(out)
                    err_log.write(err)
                
                time.sleep(0.1)




class Environs:
    def __init__(self, *args, **kwargs):
        self.cluster = kwargs.get('cluster')
        self.user = kwargs.get('user')
        self.log = kwargs.get('log')
        self.environs_data = self.load_environs(ENVIRONS_JSON).get(self.cluster)
        self.popen_command = self.environs_data["popen"]
        self.threads = kwargs.get('threads')
        self.ram = kwargs.get('gb_ram')

    def id_generator(self, size=10, chars=string.ascii_uppercase + string.digits):
        return ''.join(random.choice(chars) for _ in range(size))

    def load_environs(self, environs_file):
        with open(environs_file, "r") as read_file:
            data = json.load(read_file)
        return data

    def get_processor_line(self, *args, **kwargs):
        if kwargs.get('threads') is None:
            threads = self.environs_data["resources"][kwargs.get('job')]["threads"]
        else:
            threads = kwargs.get('threads')
        if kwargs.get('gb_ram') is None:
            ram = self.environs_data["resources"][kwargs.get('job')]["ram"]
        else:
            ram = kwargs.get('ram')
        return [threads, ram]

    def generate_job(self, commands, job):
        bash_scripts=[]
        torun = len(commands)
        threads,ram = self.get_processor_line(threads = self.threads, ram = self.ram, job = job)
        for i in range(torun):
            # bash_script = {  "PBS" : "#!/bin/bash\n#PBS -N %s\n#PBS -l %s\n#PBS -j oe\n#PBS -o $PBS_O_WORKDIR/logtmp\n#PBS -A %s\ncd $PBS_O_WORKDIR\n{%s} 2>&1 | tee %s\nsed -e 's/^/[HENIPIPE] JOB: %s:\t\t/' %s >> %s\nrm %s\n" % (job_name, self.processor_line, self.user, command, log_file, job_name, log_file, self.log_name, log_file),
                            # "SLURM" : "#!/bin/bash\n#SBATCH --job-name=%s\n#SBATCH --output=tmp\n#SBATCH --error=tmp\n#SBATCH --ntasks=1\n%s\n{%s} 2>&1 | tee %s\nsed -e 's/^/[HENIPIPE] JOB: %s:\t\t/' %s >> %s\nrm %s\n" % (job_name, self.processor_line, command, log_file, job_name, log_file, self.log_name, log_file)}
            bash_script = self.assemble_script(   LOG_FILE = self.log, \
                                    MODULES = self.environs_data["resources"][job]["modules"], \
                                    TEMP_LOG_FILE = self.id_generator(), \
                                    JOB_NAME = (job + "_" + commands[i][0]), \
                                    COMMAND = commands[i][1],
                                    RAM = ram,
                                    THREADS = threads,
                                    USER = self.user,
                                    TIME = str(datetime.now().strftime("%Y-%m-%d_%H:%M:%S")))
            bash_scripts.append(bash_script)
        return bash_scripts

    def assemble_script(self, *args, **kwargs):
        script_list=[]
        order=[]
        fn_args = kwargs
        for key, value in self.environs_data["script_lines"].items():
            script_list.append(value)
            order.append(key)
        script_list = [x for _,x in sorted(zip(order,script_list))]
        lines_unparsed = [x[0] for x in script_list]
        values_to_insert = [x[1].split("|") for x in script_list]
        lines_parsed = []

        if self.cluster == 'local':
            lines_parsed = [kwargs['COMMAND']]
        else:
            global to_test
            to_test=[lines_unparsed, values_to_insert, fn_args]
            for i in range(len(lines_unparsed)):
                if values_to_insert[i][0] is "":
                    lines_parsed.append(lines_unparsed[i])
                else:
                    string=lines_unparsed[i]
                    for j in range(len(values_to_insert[i])):
                        string = re.sub("<--{0}-->".format(j), fn_args.get(values_to_insert[i][j]), string)
                    lines_parsed.append(string)
        return "\n".join(lines_parsed)


class Fastqc(SampleFactory, object):
    def __init__(self, *args, **kwargs):
        super(Fastqc, self).__init__(*args, **kwargs)
        self.bowtie_flags = kwargs.get('bowtie_flags')
        self.job = "HENIPIPE_FASTQC"
        self.commands = self.fastqc_executable()
        self.bash_scripts = self.environs.generate_job(self.commands, self.job)
    def __call__():
        pass

    def fastqc_executable(self):
        commandline=""
        command = []
        for sample in self.runsheet_data:
            fastq1=re.sub('\t', ' ', sample['fastq1'])
            fastq2=re.sub('\t', ' ', sample['fastq2'])
            JOBSTRING = self.id_generator(size=10)
            commandline = """fastqc %s %s\n""" % (fastq1, fastq2)
            command.append([sample['sample'], commandline])
        return command


class Align(SampleFactory, object):
    def __init__(self, *args, **kwargs):
        super(Align, self).__init__(*args, **kwargs)
        self.bowtie_flags = kwargs.get('bowtie_flags')
        self.job = "HENIPIPE_ALIGN"
        self.pipe = not kwargs.get('no_pipe')
        self.filter_string = self.make_filter_string(kwargs.get('filter')[0], kwargs.get('filter')[1])
        self.norm_method = kwargs.get('norm_method')
        self.commands = self.align_executable()
        self.bash_scripts = self.environs.generate_job(self.commands, self.job)
    def __call__():
        pass

    def make_filter_string(self, low, high):
        if low is None and high is None:
            return("")
        if low is not None and high is None:
            return("-fl %s" % low)
        if low is None and high is not None:
            return("-fh %s" % high)
        if low is not None and high is not None:
            return("-fl %s -fh %s" % (low, highsCYB))

    def align_executable(self):
        commandline=""
        command = []
        for sample in self.runsheet_data:
            fastq1=re.sub('\t', ',', sample['fastq1'])
            fastq2=re.sub('\t', ',', sample['fastq2'])
            JOBSTRING = self.id_generator(size=10)
            norm_bowtie_flags='--end-to-end --very-sensitive --no-overlap --no-dovetail --no-mixed --no-discordant -q --phred33 -I 10 -X 700'
            sam2bed_string = """| samTobed - -o %s %s""" % (sample['bed_out']+'tmp', self.filter_string)
            if self.pipe:
                commandline = """echo '\n[BOWTIE] Running Bowtie piped to samTobed.py for main alignment... Output:\n'\nbowtie2 %s -p %s -1 %s -2 %s -x %s %s\n""" % (self.bowtie_flags, self.threads, fastq1, fastq2, sample['fasta'], sam2bed_string)
                commandline = commandline + """\necho 'Sorting Bed...\n'\nsort -k1,1 -k2n,2n %s > %s\n""" % (sample['bed_out']+'tmp', sample['bed_out'])
                commandline = commandline + """rm %s \n""" % (sample['bed_out']+'tmp')
            else:
                commandline = """echo '\n[BOWTIE] Running Bowtie for main alignment... Output:\n'\nbowtie2 %s -p %s -1 %s -2 %s -x %s -S %s\n""" % (self.bowtie_flags, self.threads, fastq1, fastq2, sample['fasta'], sample['sam'])
                commandline = commandline + """sleep 20s \nsamtools view -bS %s -o %s\n""" % (sample['sam'], sample['bam']+'US')
                commandline = commandline + """sleep 20s \necho '\n[SAMTOOLS]... Sorting bam file %s\n'\nsamtools sort -o %s -O bam -T %s %s\n""" % (sample['bam'], sample['bam'], sample['bam']+'US', sample['bam']+'US')
                commandline = commandline + """sleep 20s \necho '\n[SAMTOOLS]... Indexing bam file %s\n'\nsamtools index %s\n""" % (sample['bam'],sample['bam'])
                commandline = commandline + """sleep 20s \necho '\n[SAMTOBED] Running samToBed for main alignment... Output:\n'\nsamTobed %s -o %s %s""" % (sample['sam'], sample['bed_out']+'tmp', self.filter_string)
                commandline = commandline + """\necho '[SORT] Sorting Bed...\n'\nsort -k1,1 -k2n,2n %s > %s\n""" % (sample['bed_out']+'tmp', sample['bed_out'])
                commandline = commandline + """rm %s \n""" % (sample['sam'])
                commandline = commandline + """rm %s %s \n""" % (sample['bed_out']+'tmp', sample['bam']+'US')
            if self.norm_method == "spike_in":
                commandline = commandline + """echo '\n[BOWTIE] Running Bowtie piped to samTobed.py for spikein... Output:\n'\nbowtie2 %s -p 4 -1 %s -2 %s -x %s | samTobed - -o %s""" % (norm_bowtie_flags, fastq1, fastq2, sample['spikein_fasta'], sample['spikein_bed_out']+'tmp')
                commandline = commandline + """\necho '[SORT] Sorting Bed for spikein...\n'\nsort -k1,1 -k2n,2n %s > %s\n""" % (sample['spikein_bed_out']+'tmp', sample['spikein_bed_out'])
                commandline = commandline + """rm %s \n""" % (sample['spikein_bed_out']+'tmp')
            command.append([sample['sample'], commandline])
        return command







class Scale(SampleFactory, object):
    def __init__(self, *args, **kwargs):
        super(Scale, self).__init__(*args, **kwargs)
        self.job = "HENIPIPE_SCALE"
        norm_method = kwargs.get('norm_method')
        #self.processor_line = self.norm_processor_line()
        self.norm_values = self.get_norm_values(method = norm_method)
        self.commands = self.norm_executable()
        self.bash_scripts = self.environs.generate_job(self.commands, self.job)
    def __call__():
        pass

    def get_norm_values(self, method):
        if method=="read_count":
            for sample in self.runsheet_data:
                count = 0
                # for line in open(sample['bed_out']).xreadlines(): count += 1
                for line in open(sample['bed_out']): count += 1
                sample['scale_factor'] = (1/float(count))*float(10**7)
        if method=="coverage":
            for sample in self.runsheet_data:
                count = 0
                with open(sample['bed_out']) as infile:
                    for line in infile:
                        count += int(line.split("\t")[4])
                sample['scale_factor'] = (1/float(count))*float(10**10)
        if method=="spike_in":
            for sample in self.runsheet_data:
                count = 0
                for line in open(sample['bed_out']): count += 1
                ncount = 0
                for line in open(sample['spikein_bed_out']): ncount += 1
                sample['scale_factor'] = (1/float(ncount))
        return

    def norm_executable(self):
        command = []
        for sample in self.runsheet_data:
            JOBSTRING = self.id_generator(size=10)
            commandline = """genomeCoverageBed -bg -i %s -g %s -scale %s -trackline | pyWriter - %s\n""" % (sample['bed_out'], sample['genome_sizes'], sample['scale_factor'], sample['bedgraph'])
            command.append([sample['sample'], commandline])
        return command






class SEACR(SampleFactory, object):
    def __init__(self, *args, **kwargs):
        super(SEACR, self).__init__(*args, **kwargs)
        self.job = "HENIPIPE_SEACR"
        self.method = kwargs.get('stringency')
        self.norm = kwargs.get('norm')
        self.runsheet_data = self.SEACR_match()
        self.processor_line = self.SEACR_processor_line()
        self.commands = self.SEACR_executable()
        self.bash_scripts = self.environs.generate_job(self.commands, self.job)
    def __call__():
        pass

    def SEACR_match(self):
        sk = [i.get('SEACR_key') for i in self.runsheet_data]
        controls_b = [bool(re.search(r'._CONTROL$', i)) for i in sk]
        controls = list(compress(self.runsheet_data, controls_b))
        samples_b = [not i for i in controls_b]
        samples = list(compress(self.runsheet_data, samples_b))
        for sample in samples:
            control_name = sample.get('SEACR_key')+"_CONTROL"
            try:
                control_bed = next(item for item in controls if item["SEACR_key"] == control_name).get('bedgraph')
            except StopIteration:
                raise ValueError("Could not find matching files, make sure it is in the runsheet and/or your aren't selecting it out with the select flag")
            sample.update( {'SEACR_in' : sample.get('bedgraph')})
            sample.update( {'SEACR_control' : control_bed})
        return samples

    def SEACR_executable(self):
        commandline=""
        command = []
        for sample in self.runsheet_data:
            JOBSTRING = self.id_generator(size=10)
            commandline = """echo '\n[SEACR] Running SEACR... Output:\n'\nbash %s %s %s %s %s %s\n""" % (SEACR_SCRIPT, sample['SEACR_in'], sample['SEACR_control'], self.norm, self.method, sample['SEACR_out'])
            command.append([sample['sample'], commandline])
        return command

    def SEACR_processor_line(self):
        if self.cluster=="PBS":
            return """select=1:mem=8GB:ncpus=2"""
        if self.cluster=="SLURM":
            return ''





class Merge(SampleFactory, object):
    def __init__(self, *args, **kwargs):
        super(Merge, self).__init__(*args, **kwargs)
        self.job = "HENIPIPE_MERGE"
        self.out = kwargs.get('out')
        self.to_process = self.Merge_match()
        self.processor_line = self.Merge_processor_line()
        self.commands = self.Merge_executable()
        self.bash_scripts = self.environs.generate_job(self.commands, self.job)
    def __call__():
        pass

    def Merge_match(self):
        key_data = [i.get("MERGE_key") for i in self.runsheet_data]
        bg_data = [i.get("bedgraph") for i in self.runsheet_data]
        merge_dict = dict.fromkeys(key_data, "NotFound")
        samples = []
        for key in merge_dict.keys():
            samples.append({    "sample" : key,
                                "files_to_merge": list(compress(bg_data, is_in(key, key_data))),})
        return(samples)


    def Merge_executable(self):
        commandline=""
        command = []
        for item in self.to_process:
            seperator = ' '
            nfiles = len(item.get("files_to_merge"))
            bedgraph_line = seperator.join(item.get("files_to_merge"))
            bedgraph_out=str(os.path.join(self.out, item.get("sample")))+"_merged.bedgraph"
            JOBSTRING = self.id_generator(size=10)
            #commandline = """echo '\n[MERGE] Merging bedgraphs:\n%s'\nbedtools unionbedg -i %s | awk '{sum=0; for (col=4; col<=NF; col++) sum += $col; print $0"\t"sum/(NF-4+1); }' > %s\nsleep 10\ncut -d$'\t' -f1-3,%s %s > %s\nrm %s\n""" % (bedgraph_line, bedgraph_line, bedgraph_out+'temp', (nfiles+4), bedgraph_out+'temp', bedgraph_out, bedgraph_out+'temp')
            commandline = """echo '\n[MERGE] Merging bedgraphs:\n%s'\nbedtools unionbedg -i %s | awk '{sum=0; for (col=4; col<=NF; col++) sum += $col; print $0"\t"sum/(NF-4+1); }' > %s\n""" % (bedgraph_line, bedgraph_line, bedgraph_out)
            command.append([item['sample'], commandline])
        return command


    def Merge_processor_line(self):
        if self.cluster=="PBS":
            return """select=1:mem=8GB:ncpus=2"""
        if self.cluster=="SLURM":
            return ''






class MACS2(SampleFactory, object):
    def __init__(self, *args, **kwargs):
        super(MACS2, self).__init__(*args, **kwargs)
        self.job = "HENIPIPE_MACS2"
        #self.merged = kwargs.get('merged')
        self.out = kwargs.get('out')
        self.norm = kwargs.get('norm')
        self.runsheet_data = self.MACS2_match()
        self.processor_line = self.MACS2_processor_line()
        self.commands = self.MACS2_executable()
        self.script = self.environs.generate_job(self,commands, self.job)
    def __call__():
        pass

    def MACS2_match(self):
        desired_samples = self.runsheet_data
        #desired_samples = parsed_runsheet
        sample_key = [i.get("sample") for i in desired_samples]
        biomatch_data = [i.get("MACS2_key") for i in desired_samples]
        abmatch_data = [i.get("SEACR_key") for i in desired_samples]
        unique_keys = unique(sample_key)
        run_list = []
        for key in unique_keys:
            #find out if file is bio sample or control or ab sample or control by searching lists of the two keys
            #print(key)
            biomatch_key = [biomatch_data[i] for i in which(key, sample_key)]
            is_biomatch_control = [bool(re.search(r'._CONTROL$', i)) for i in biomatch_key]
            abmatch_key = [abmatch_data[i] for i in which(key, sample_key)]
            is_abmatch_control = [bool(re.search(r'._CONTROL$', i)) for i in abmatch_key]
            is_biomatch_control = all_the_same(is_biomatch_control)
            is_abmatch_control = all_the_same(is_abmatch_control)
            if type(is_abmatch_control) is str:
                raise ValueError("Some discrepency between merge_key and MACS2_key ")
            if type(is_biomatch_control) is str:
                raise ValueError("Some discrepency between merge_key and MACS2_key ")
            if not is_abmatch_control and not is_biomatch_control:
                #treatment bed is just key
                treatment_bed = get_key_from_dict_list(desired_samples, {"sample":key}, 'bed_out')
                #for all non-controls we will output a list of relevant files to process using macs2
                #first find biomatch control
                biomatch_control_key = biomatch_key[0]+"_CONTROL"
                #get bed of this control
                control_bed = get_key_from_dict_list(desired_samples, {"sample":sample_key[which(biomatch_control_key, biomatch_data)[0]]}, 'bed_out')
                #get bed of antibody control for treatment
                abmatch_control_key = abmatch_key[0]+"_CONTROL"
                treatment_abcontrol_bed = get_key_from_dict_list(desired_samples, {"sample":sample_key[which(abmatch_control_key, abmatch_data)[0]]}, 'bed_out')
                #get bed of antibody control for control
                control_control_key = get_key_from_dict_list(desired_samples, {"sample":sample_key[which(biomatch_control_key, biomatch_data)[0]]}, 'SEACR_key')+"_CONTROL"
                control_abcontrol_bed = get_key_from_dict_list(desired_samples, {"sample":sample_key[which(control_control_key, abmatch_data)[0]]}, 'bed_out')
                run_list.append({   "MACS2DIFF_treatment": key,
                                    "MACS2CP_treat_sample": treatment_bed,
                                    "MACS2CP_treat_control": treatment_abcontrol_bed,
                                    "MACS2DIFF_control": sample_key[which(biomatch_control_key, biomatch_data)[0]],
                                    "MACS2CP_control_sample": control_bed,
                                    "MACS2CP_control_control": control_abcontrol_bed,
                                    "sample": key})
        return(run_list)






    def MACS2_executable(self):
        commandline=""
        command = []
        for item in self.runsheet_data:
            JOBSTRING = self.id_generator(size=10)
            treat_p = os.path.join(self.out, item["MACS2DIFF_treatment"])
            cont_p = os.path.join(self.out, item["MACS2DIFF_control"])
            if self.cluster=="SLURM":
                modules = """\nsource /app/Lmod/lmod/lmod/init/bash\nmodule load MACS2"""
            else:
                modules = """\n"""
            commandline = """echo '\n[MACS2] Running MACS2 callpeak on sample... Output:\n'\nmacs2 callpeak -B -t %s -c %s -f BEDPE -g hs --nomodel --extsize 147 --outdir %s -n %s\n""" % (item["MACS2CP_treat_sample"], item["MACS2CP_treat_control"], self.out, item["MACS2DIFF_treatment"])
            commandline = commandline + """echo '\n[MACS2] Getting depth of sample... Output:\n'\nstr1=$(egrep "fragments after filtering in control" %s_peaks.xls | cut -d ":" -f2)\n""" % (treat_p)
            commandline = commandline + """echo '\n[MACS2] Running MACS2 callpeak on control... Output:\n'\nmacs2 callpeak -B -t %s -c %s -f BEDPE -g hs --nomodel --extsize 147 --outdir %s -n %s\n""" % (item["MACS2CP_control_sample"], item["MACS2CP_control_control"], self.out, item["MACS2DIFF_control"])
            commandline = commandline + """echo '\n[MACS2] Getting depth of sample... Output:\n'\nstr2=$(egrep "fragments after filtering in control" %s_peaks.xls | cut -d ":" -f2)\n""" % (cont_p)
            commandline = commandline + """echo '\n[MACS2] Running MACS2 bdgdiff... Output:\n'\nmacs2 bdgdiff --t1 %s_treat_pileup.bdg --c1 %s_control_lambda.bdg --t2 %s_treat_pileup.bdg --c2 %s_control_lambda.bdg --d1 $str1 --d2 $str2 -g 60 -l 147 --o-prefix %s --outdir %s\n""" % (treat_p, treat_p, cont_p, cont_p, item["MACS2DIFF_treatment"]+"_v_"+item["MACS2DIFF_control"], self.out)
            commandline = modules + commandline
            command.append(commandline)
        return command

    def MACS2_processor_line(self):
        if self.cluster=="PBS":
            return """select=1:mem=8GB:ncpus=2"""
        if self.cluster=="SLURM":
            return ''



class AUC(SampleFactory, object):
    def __init__(self, *args, **kwargs):
        super(AUC, self).__init__(*args, **kwargs)
        self.job = "HENIPIPE_AUC"
        self.out = kwargs.get('out')
        self.norm = kwargs.get('norm')
        self.pipe = not kwargs.get('no_pipe')
        self.method = kwargs.get('stringency')
        self.runsheet_data = self.AUC_match()
        self.processor_line = self.AUC_processor_line()
        self.commands = self.AUC_executable()
        self.script = self.environs.generate_job(self,commands, self.job)
    def __call__():
        pass

    def AUC_match(self):
        desired_samples = self.runsheet_data
        sample_key = [i.get("sample") for i in desired_samples]
        biomatch_data = [i.get("MACS2_key") for i in desired_samples]
        abmatch_data = [i.get("SEACR_key") for i in desired_samples]
        unique_keys = unique(sample_key)
        run_list = []
        for key in unique_keys:
            #find out if file is bio sample or control or ab sample or control by searching lists of the two keys
            biomatch_key = [biomatch_data[i] for i in which(key, sample_key)]
            is_biomatch_control = [bool(re.search(r'._CONTROL$', i)) for i in biomatch_key]
            abmatch_key = [abmatch_data[i] for i in which(key, sample_key)]
            is_abmatch_control = [bool(re.search(r'._CONTROL$', i)) for i in abmatch_key]
            is_biomatch_control = all_the_same(is_biomatch_control)
            is_abmatch_control = all_the_same(is_abmatch_control)
            if type(is_abmatch_control) is str:
                raise ValueError("Some discrepency between merge_key and MACS2_key ")
            if type(is_biomatch_control) is str:
                raise ValueError("Some discrepency between merge_key and MACS2_key ")
            if not is_abmatch_control and not is_biomatch_control:
                #treatment bed is just key
                treatment_bed = get_key_from_dict_list(desired_samples, {"sample":key}, 'bedgraph')
                #for all non-controls we will output a list of relevant files to process using macs2
                #first find biomatch control
                biomatch_control_key = biomatch_key[0]+"_CONTROL"
                #get bed of this control
                control_bed = get_key_from_dict_list(desired_samples, {"sample":sample_key[which(biomatch_control_key, biomatch_data)[0]]}, 'bedgraph')
                #get bed of antibody control for treatment
                abmatch_control_key = abmatch_key[0]+"_CONTROL"
                treatment_abcontrol_bed = get_key_from_dict_list(desired_samples, {"sample":sample_key[which(abmatch_control_key, abmatch_data)[0]]}, 'bedgraph')
                #get bed of antibody control for control
                control_control_key = get_key_from_dict_list(desired_samples, {"sample":sample_key[which(biomatch_control_key, biomatch_data)[0]]}, 'SEACR_key')+"_CONTROL"
                control_abcontrol_bed = get_key_from_dict_list(desired_samples, {"sample":sample_key[which(control_control_key, abmatch_data)[0]]}, 'bedgraph')
                run_list.append({   "AUC_DIFF_treatment": key,
                                    "AUC_CP_treat_sample": treatment_bed,
                                    "AUC_CP_treat_control": treatment_abcontrol_bed,
                                    "AUC_DIFF_control": sample_key[which(biomatch_control_key, biomatch_data)[0]],
                                    "AUC_CP_control_sample": control_bed,
                                    "AUC_CP_control_control": control_abcontrol_bed,
                                    "sample": key})
        return(run_list)

    def AUC_executable(self):
        commandline=""
        command = []
        for item in self.runsheet_data:
            self.files_rm = []
            JOBSTRING = self.id_generator(size=10)
            treat_comb = os.path.join(self.out, (item["AUC_DIFF_treatment"]+"_"+item["AUC_DIFF_control"]+"_treatment_combined.bedgraph"))
            cont_comb = os.path.join(self.out, (item["AUC_DIFF_treatment"]+"_"+item["AUC_DIFF_control"]+"_cont_combined.bedgraph"))
            seacr_merge_prefix = os.path.join(self.out, (item["AUC_DIFF_treatment"]+"_"+item["AUC_DIFF_control"]+"_SEACR"))
            peakfile = os.path.join(self.out, (item["AUC_DIFF_treatment"]+"_"+item["AUC_DIFF_control"]+"_SEACR.")+self.method+".bed")
            out_file = os.path.join(self.out, (item["AUC_DIFF_treatment"]+"_"+item["AUC_DIFF_control"]+"_AUC.bed"))
            cp_treat_out = os.path.join(self.out, (os.path.basename(item["AUC_CP_treat_sample"])+'.bz'))
            cp_cont_out = os.path.join(self.out, (os.path.basename(item["AUC_CP_control_sample"])+'.bz'))
            self.files_rm.extend((cp_treat_out, cp_cont_out, treat_comb, cont_comb, cp_treat_out+".tbi", cp_cont_out+".tbi"))
            if self.cluster=="SLURM":
                modules = """\nsource /app/Lmod/lmod/lmod/init/bash\nmodule load bedtools\nmodule load R\nmodule load htslib/1.9\n"""
            else:
                modules = """\nmodule load bedtools\nmodule load R\nmodule load htslib/1.9\n"""
            commandline = """echo '\n[AUC] Merging sample bedgraphs for aggregated peak call...'\nbedtools unionbedg -i %s %s | awk '{sum=0; for (col=4; col<=NF; col++) sum += $col; print $1"\t"$2"\t"$3"\t"sum; }' > %s\n""" % (item["AUC_CP_treat_sample"], item["AUC_CP_control_sample"], treat_comb)
            commandline = commandline + """echo '\n[AUC] Merging control bedgraphs for aggregated peak call...'\nbedtools unionbedg -i %s %s | awk '{sum=0; for (col=4; col<=NF; col++) sum += $col; print $1"\t"$2"\t"$3"\t"sum; }' > %s\n""" % (item["AUC_CP_treat_control"], item["AUC_CP_control_control"], cont_comb)
            commandline = commandline + """echo '\n[AUC] Running SEACR on merged sample data... Output:\n'\nbash %s %s %s %s %s %s\n""" % (SEACR_SCRIPT, treat_comb, cont_comb, self.norm, self.method, seacr_merge_prefix)
            commandline = commandline + """echo '\n[AUC] Making Tabix files... \n'\nbgzip -c %s > %s\n""" % (item["AUC_CP_treat_sample"], cp_treat_out)
            commandline = commandline + """bgzip -c %s > %s\n""" % (item["AUC_CP_control_sample"], cp_cont_out)
            commandline = commandline + """tabix -S 1 -p bed %s\n""" % (cp_treat_out)
            commandline = commandline + """tabix -S 1 -p bed %s\n""" % (cp_cont_out)
            commandline = commandline + """auc -o %s -p %s %s %s\n""" % (out_file, peakfile, cp_treat_out, cp_cont_out)
            if self.pipe: commandline = commandline + """echo '\n[AUC] Removing files'\nrm {0}\n""".format(" ".join(self.files_rm))
            commandline = modules + commandline
            command.append(commandline)
        return command

    def AUC_processor_line(self):
        if self.cluster=="PBS":
            return """select=1:mem=4GB:ncpus=1"""
        if self.cluster=="SLURM":
            return ''



def get_key_from_dict_list(list_of_dicts, dict_in, key_out):
    #this function takes a list of dicts, returns the value associated with key_out for the element in the list_of_dicts which has the key:value combo given in dict_in
    dict_out = next(item for item in list_of_dicts if item[list(dict_in.keys())[0]] == dict_in.get(list(dict_in.keys())[0]))
    return dict_out.get(key_out)

def which(string, list):
    return [i for i, x in enumerate(list) if x == string]

def unique_indices(list):
    result = {}
    for key, val in zip(keys, vals):
        if key not in result:
            result[key] = 0
        result[key] += val
    return(result)


def unique(list1):
    # function to get unique values
    # intilize a null list
    unique_list = []
    # traverse for all elements
    for x in list1:
        # check if exists in unique_list or not
        if x not in unique_list:
            unique_list.append(x)
    return(unique_list)

def convert_windows_newlines(file_name):
    """`
    Helper function for converting windows newlines in a file as a preprocessing step in case users make samplesheet
    in excel.

    Args:
    file_name (str): name of file to convert to unix newlines.

    """
    # Read in the file in entirety first
    with open(file_name) as input:
        file_text = input.read()

    # Replace newlines produced by excel
    file_text = file_text.replace('\r', '\n')

    # Write the new output back to the file
    with open(file_name, 'w') as output:
        output.write(file_text)

def is_in(string, list):
    outlist=[]
    #This function returns a list of booleans for matching a string in a list of strings
    for i in list:
        if i == string:
            outlist.append(True)
        else:
            outlist.append(False)
    return(outlist)

def reduce_concat(x, sep=""):
    return functools.reduce(lambda x, y: str(x) + sep + str(y), x)

def find_fastq_mate(dir, sample_flag=None, full_name=True):
    fastqs=[]
    fastq1=[]
    fastq2=[]
    for file in os.listdir(dir):
        if file.endswith(".fastq.gz"):
            fastqs.extend([file])
            if "_R1_" in file:
                fastq1.extend([file])
            if "_R2_" in file:
                fastq2.extend([file])
    fastq1_mate=[]
    for fastq in fastq1:
        #check if present
        put_R2=re.sub('_R1_', '_R2_', fastq)
        try:
            fastq1_mate.extend([fastq2[fastq2.index(put_R2)]])
        except ValueError:
            raise ValueError("Could not find matching file: "+put_R2+" for: "+fastq+" in "+dir)
    if full_name:
            for i in range(len(fastq1)):
                fastq1[i] = os.path.join(dir, fastq1[i])
                fastq1_mate[i] = os.path.join(dir, fastq1_mate[i])
    keys=['directory_long', 'directory_short','fastq1', 'fastq2', 'has_fastq']
    #values=[dir, os.path.basename(dir), "\t".join(os.path.join(dir, fastq1)), "\t".join(os.path.join(dir, fastq1_mate)), len(fastq1)>0]
    values=[dir, os.path.basename(dir), "\t".join(fastq1), "\t".join(fastq1_mate), len(fastq1)>0]
    return(dict(zip(keys, values)))

def all_the_same(list_in):
    #this function takes a list of booleans and characterizes them as all true, all false or mixed.
    if all(list_in) is True:
        return(True)
    elif all([not i for i in list_in]) is True:
        return(False)
    else:
        return("mixed")

def find_colnames(runsheet, header=True):
    if header==True:
        with open(runsheet, 'r') as f:
            reader = csv.reader(f)
            return(next(reader))            # read header


def load_genomes(genomes_file):
    with open(genomes_file, "r") as read_file:
        genome_data = json.load(read_file)
    return genome_data

def make_runsheet(folder, sample_flag, genome_key, output="./henipipeout", fasta=None, spikein_fasta=None, genome_sizes=None, no_pipe = True):
    genome_data = load_genomes(GENOMES_JSON).get(genome_key)
    ddir=[x[0] for x in os.walk(folder)]
    dat=list(map(find_fastq_mate, ddir))
    good_dat = [i for i in dat if i.get('has_fastq') is True]
    good_dat = [i for i in good_dat if re.compile(r'.*'+sample_flag).search(i.get('directory_short'))]
    for i in good_dat:
        i.update({'sample': i.get('directory_short'), \
            'bed_out': os.path.join(output, i.get('directory_short')+".bed"), \
            'spikein_bed_out': os.path.join(output, i.get('directory_short')+"_spikein.bed"), \
            'bedgraph': os.path.join(output, i.get('directory_short')+".bedgraph"), \
            'SEACR_key': i.get('directory_short'), \
            'MERGE_key': i.get('directory_short'), \
            'SEACR_out': os.path.join(output, i.get('directory_short')+"_SEACR"), \
            'fasta': genome_data.get('fasta'), 'spikein_fasta': genome_data.get('spikein_fasta'), 'genome_sizes':  genome_data.get('genome_sizes')})
        if no_pipe:
            i.update({'sam': os.path.join(output, i.get('directory_short')+".sam"), \
                'bam': os.path.join(output, i.get('directory_short')+".bam")})
    keys = ["sample", "SEACR_key", "MERGE_key", "fasta", "spikein_fasta", "genome_sizes", "fastq1", "fastq2", "bed_out", "spikein_bed_out", "bedgraph", "SEACR_out"]
    if no_pipe:
        keys.append("sam")
        keys.append("bam")
    with open(os.path.join(output, 'runsheet.csv'), 'w') as output_file:
        dict_writer = csv.DictWriter(output_file, fieldnames = keys, extrasaction='ignore')
        dict_writer.writeheader()
        dict_writer.writerows(good_dat)

def parse_runsheet(runsheet, header=True, colnames=None):
    """
    Helper function for parsing runsheets provided by user.
    Args:
        runsheet (file): file object to runsheet.
    Yields:
        Each line of the run as a dict by column name.
        Columns: as listed in csv
    """
    if header:
        columns = find_colnames(runsheet, header=header)
    if header==False and colnames is not None:
        columns = colnames
    if header==False and colnames is None:
        raise ValueError('No header indicated in runsheet and no colnames provided')

    entries = []
    with open(runsheet, 'r') as f:
        reader = csv.reader(f)
        if header==True:
            next(reader)            # skip header
        for line in f:
            entries = [entry.strip() for entry in line.strip().split(',')]
            if len(entries) != len(columns):
                raise ValueError('Sample sheet does not match expected columns. Expects: %s' % ','.join(columns))
            entry_dict = dict(zip(columns, entries))
            yield entry_dict

def check_runsheet_parameter(runsheet, parameter, verbose=False):
    for i in runsheet:
        if i.get(parameter) is None:
            if verbose: print("no data for paramter "+parameter)
            return 0
        if i.get(parameter) is "":
            if verbose: print("header present, but no or incomplete data for "+parameter)
            return 0
    if verbose: print("runsheet_okay_for_parameter_"+parameter)
    return 1

def check_runsheet(args, runsheet, verbose=False):
    required_args = ['sample', 'fastq1', 'fastq2', 'bed_out', 'bedgraph', 'genome_sizes']
    if args.norm_method == "spike_in":
        required_args.extend(['spikein_fasta', 'spikein_bed_out'])
    if args.job == "SEACR":
        required_args.extend(['SEACR_key', 'SEACR_out'])
    if args.job == "MERGE":
        required_args.extend(['MERGE_key'])
    has_data = []
    for i in required_args:
        has_data.append(check_runsheet_parameter(runsheet, i, verbose = verbose))
    has_data = dict(zip(required_args, has_data))
    listOfKeys = getKeysByValues(has_data, [0])
    error_string = "Data not found for the following required runsheet columns: "
    if len(listOfKeys) != 0:
        for key in listOfKeys:
            error_string = error_string + key + ', '
        raise ValueError(error_string)
    return

def getKeysByValues(dictOfElements, listOfValues):
    listOfKeys = list()
    listOfItems = dictOfElements.items()
    for item  in listOfItems:
        if item[1] in listOfValues:
            listOfKeys.append(item[0])
    return  listOfKeys


def parse_range_list(rl):
    def collapse_range(ranges):
        end = None
        for value in ranges:
            yield range(max(end, value.start), max(value.stop, end)) if end else value
            end = max(end, value.stop) if end else value.stop
    def split_range(value):
        value = value.split(':')
        for val, prev in zip(value, chain((None,), value)):
            if val != '':
                val = int(val)
                if prev == '':
                    val *= -1
                yield val
    def parse_range(r):
        parts = list(split_range(r.strip()))
        if len(parts) == 0:
            return range(0, 0)
        elif len(parts) > 2:
            raise ValueError("Invalid range: {}".format(r))
        return range(parts[0], parts[-1] + 1)
    ranges = sorted(set(map(parse_range, rl.split(","))), key=lambda x: (x.start, x.stop))
    return chain.from_iterable(collapse_range(ranges))



#!/usr/bin/python

import sys
try:
    from collections import OrderedDict
except ImportError: #python 2.6 or 3.6+
    if sys.version_info >= (3,6):
        OrderedDict = dict
    else:
        from ordereddict import OrderedDict

import os
from itertools import groupby
from subprocess import Popen, PIPE
from io import TextIOWrapper
import re
import argparse
from six import PY3, string_types

try:
    from multiprocessing.dummy.connection import Connection
except ImportError: #python2
    from _multiprocessing import Connection

__version__ = '0.1'

class DefaultOrderedDict(OrderedDict):
    def __init__(self, default, items=[]):
        super(DefaultOrderedDict, self).__init__(items)
        self._default = default

    def __missing__(self, key):
        self[key] = value = self._default()
        return value


class GenomicOrder(object):
    def __gt__(self, other):
        if self.rname != other.rname:
            return self.rname > other.rname
        return self.pos > other.pos

    def __lt__(self, other):
        if self.rname != other.rname:
            return self.rname < other.rname
        return self.pos < other.pos

    def __eq__(self, other):
        return self.rname == other.rname and self.pos == other.pos


class Reader(object):
    """ Read SAM/BAM format file as an iterable. """
    def __init__(self, f, regions=False, kind=None, samtools_path="samtools"):
        ext = None
        self.samtools_path = samtools_path
        self.spool = None  # use this to catch alignment during reader scraping
        self.type = 'sam'
        try:
            self._f_name = f.name
            _, ext = os.path.splitext(f.name)
            if f.name == '<stdin>':  # stdin stream
                self._sam_init(f)
            elif (ext is not None and ext.lower()) == '.bam' or (kind is not None and kind.lower() == 'bam'):
                self._bam_init(f, regions)
                self.type = 'bam'
            elif (ext is not None and ext.lower()) == '.sam' or (kind is not None and kind.lower() == 'sam'):
                self._sam_init(f)
            else:
                self._sam_init(f)
            if (regions and (ext is not None and ext.lower() != '.bam') and kind is None) or (regions and kind is not None and kind.lower() != 'bam'):
                self.__exit__()
                raise ValueError("Region support requires bam file.")
        except AttributeError:
            self._f_name = None
            if isinstance(f, Connection):
                self._pipe_init(f)
            else:
                self._sam_init(f)

    def _pipe_init(self, f):
        header = []
        for line in iter(f.recv, ''):
            if line[0] == '@':
                header.append(line.rstrip('\n\r'))
            else:
                self.spool = line
                break
        self.header_as_dict(header)
        self.f = iter(f.recv, '')
        self._conn = 'pipe'

    def _sam_init(self, f):
        header = []
        self.f = f
        for line in self.f:
            if line[0] == '@':
                header.append(line.rstrip('\n\r'))
            else:
                self.spool = line
                break
        self.header_as_dict(header)
        self._conn = 'file'

    def _bam_init(self, f, regions):
        pline = [self.samtools_path, 'view', '-H', f.name]
        try:
            p = Popen(pline, bufsize=-1, stdout=PIPE,
                      stderr=PIPE)
        except OSError:
            raise OSError('Samtools must be installed for BAM file support!\n')
        self.header_as_dict([line.decode('utf-8').rstrip('\n\r') for line in p.stdout])
        p.wait()
        if regions:
            try:
                open(''.join([f.name, '.bai']))
            except EnvironmentError:
                sys.stderr.write("BAM index not found. Attempting to index file.\n")
                index_p = Popen([self.samtools_path, 'index', f.name], stdout=PIPE, stderr=PIPE)

                _, err = index_p.communicate()
                if index_p.returncode > 0 or re.search("fail", str(err)):
                    raise OSError("Indexing failed. Is the BAM file sorted?\n")
                else:
                    sys.stderr.write("Index created successfully.\n")
            pline = [self.samtools_path, 'view', f.name, regions]
        else:
            pline = [self.samtools_path, 'view', f.name]
        self.p = Popen(pline, bufsize=-1, stdout=PIPE,
                  stderr=PIPE)
        if PY3:
            self.f = TextIOWrapper(self.p.stdout)
        else:
            self.f = self.p.stdout

        self._conn = 'proc'

    def next(self):
        """ Returns the next :class:`.Sam` object """
        try:
            if self.spool:  # this will be the first alignment in a SAM file or stream
                line = self.spool.rstrip('\n\r')
                self.spool = None
            else:
                line = next(self.f).rstrip('\n\r')
            if line == '':
                raise StopIteration
            fields = line.split('\t')
            required = fields[:11]
            tags = fields[11:]
            return Sam(*required, tags=tags)
        except StopIteration:
            raise StopIteration

    def __next__(self):
        return self.next()

    def __iter__(self):
        return self

    def __len__(self):
        """ Returns the number of reads in an indexed BAM file.
        Not implemented for SAM files. """
        if self.type != 'bam':
            raise NotImplementedError("len(Reader) is only implemented for BAM files.")
        elif self.type == 'bam':
            return sum(bam_read_count(self._f_name, self.samtools_path))

    def subsample(self, n):
        """ Returns an interator that draws every nth read from
        the input file. Returns :class:`.Sam`. """
        for i, line in enumerate(self.f):
            if i % n == 0:
                fields = line.split('\t')
                required = fields[:11]
                tags = fields[11:]
                yield Sam(*required, tags=tags)

    def header_as_dict(self, header):
        """ Parse the header list and return a nested dictionary. """
        self.header = DefaultOrderedDict(OrderedDict)
        for line in header:
            line = line.split('\t')
            key, fields = (line[0], line[1:])
            try:
                self.header[key][fields[0]] = fields[1:]
            except IndexError:
                self.header[key][fields[0]] = ['']

    @property
    def seqs(self):
        """ Return just the sequence names from the @SQ library as a generator. """
        for key in self.header['@SQ'].keys():
            yield key.split(':')[1]

    def tile_genome(self, width):
        """ Return a generator of UCSC-style regions tiling ``width``. """
        assert isinstance(width, int)
        for k, v in self.header['@SQ'].items():
            rname = k.split(':')[1]
            seqlength = v[0].split(':')[1]
            for region in tile_region(rname, 1, int(seqlength), width):
                yield region

    def close(self):
        self.__exit__()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        if self._conn == 'file':
            self.f.close()
        if self._conn == 'proc':
            self.f.close()
            self.p.terminate()


class Writer(object):
    """ Write SAM/BAM format file from :class:`.Sam` objects. """
    def __init__(self, f, header=None):
        try:
            _, ext = os.path.splitext(f.name)
            if ext == '.bam':
                raise NotImplementedError('Bam writing support is not implemented.\n') # why not just pipe to samtools?
        except AttributeError:
            pass
        self.file = f
        if header is not None:
            self.header = DefaultOrderedDict(OrderedDict)
            self._merge_header(header)
        else:
            self.header = DefaultOrderedDict(OrderedDict)
            self.header['@HD']['VN:1.0'] = ['SO:unknown']
        self._header_dict_format()

    def _merge_header(self, header):
        for key, values in header.items():
            for k, v in values.items():
                self.header[key][k] = v

    def _header_dict_format(self):
        for key, value in self.header.items():
            for k, v in value.items():
                tags = '\t'.join(v)
                self.file.write('{key}\t{k}\t{tags}\n'.format(**locals()))

    def write(self, sam):
        """ Write the string representation of the ``sam`` :class:`.Sam` object. """
        self.file.write(str(sam))

    def close(self):
        self.__exit__()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.file.close()


class Sam(GenomicOrder):
    """ Object representation of a SAM entry. """
    # https://github.com/samtools/hts-specs/blob/da805be01e2ceaaa69fdde9f33c5377bf9ee6369/SAMv1.tex#L383
    # operations that consume the reference
    _cigar_ref = set(('M', 'D', 'N', '=', 'X', 'EQ'))
    # operations that consume the query
    _cigar_query = set(('M', 'I', 'S', '=', 'X', 'EQ'))
    # operations that do not represent an alignment
    _cigar_no_align = set(('H', 'P'))
    _valid_cigar = _cigar_ref | _cigar_query | _cigar_no_align
    # operations that can be represented as aligned to the reference
    _cigar_align = _cigar_ref & _cigar_query
    # operations that only consume the reference
    _cigar_ref_only = _cigar_ref - _cigar_align
    # operations that only consume the query
    _cigar_query_only = _cigar_query - _cigar_align

    def __init__(self, qname='', flag=4, rname='*', pos=0, mapq=255, cigar='*', rnext='*', pnext=0, tlen=0, seq='*', qual='*', tags=[]):
        self.qname = qname
        self.flag = int(flag)
        self.rname = rname
        self.pos = int(pos)
        self.mapq = int(mapq)
        self.cigar = cigar
        self.rnext = rnext
        self.pnext = int(pnext)
        self.tlen = int(tlen)
        self.seq = seq
        self.qual = qual
        self._tags = tags
        self._cache = dict()

    def __str__(self):
        """ Returns the string representation of a SAM entry. Correspondes to one line
        in the on-disk format of a SAM file. """
        if self.tags:
            tag_fields = '\t'.join([encode_tag(tag, self.tags[tag]) for tag in sorted(self.tags.keys())])
        else:
            tag_fields = '\t'.join(self._tags)
        return '{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\t{7}\t{8}\t{9}\t{10}\t{11}\n'.format(self.qname,
                                                                                     str(self.flag),
                                                                                     self.rname,
                                                                                     str(self.pos),
                                                                                     str(self.mapq),
                                                                                     self.cigar,
                                                                                     self.rnext,
                                                                                     str(self.pnext),
                                                                                     str(self.tlen),
                                                                                     self.seq,
                                                                                     self.qual,
                                                                                     tag_fields)
    def __repr__(self):
        return "Sam({0}:{1}:{2})".format(self.rname, self.pos, self.qname)

    def __len__(self):
        """ Returns the length of the portion of ``self.seq`` aligned to the reference. Unaligned reads will
        have len() == 0. Insertions (I) and soft-clipped portions (S) will not contribute to the aligned length. 

        >>> x = Sam(cigar='8M2I4M1D3M4S')
        >>> len(x)
        16
        """
        return sum(c[0] for c in self.cigars if c[1] in self._cigar_ref)

    def __getitem__(self, tag):
        """ Retreives the SAM tag named "tag" as a tuple: (tag_name, data). The
        data type of the tag is interpreted as the proper Python object type.

        >>> x = Sam(tags=['NM:i:0', 'ZZ:Z:xyz'])
        >>> x['NM']
        0
        >>> x['ZZ']
        'xyz'
        """
        return self.tags[tag]

    def __setitem__(self, tag, data):
        """ Stores the SAM tag named "tag" with the value "data". The
        data type of the tag is interpreted from the Python object type.

        >>> x = Sam(tags=[])
        >>> x['NM'] = 0
        >>> x['NM']
        0
        """
        self.tags[tag] = data

    def index_of(self, pos):
        """ Return the relative index within the alignment from a genomic position 'pos' """
        i = pos - self.pos
        if i >= 0:
            return i
        else:
            raise IndexError("Position {0:n} not in {1}.".format(pos, self.qname))

    def get(self, key, default_value):
        try:
            return self[key]
        except KeyError:
            return default_value

    def cigar_split(self):
        # https://github.com/brentp/bwa-meth
        if self.cigar == "*":
            yield (0, None)
            raise StopIteration
        cig_iter = groupby(self.cigar, lambda c: c.isdigit())
        for _, n in cig_iter:
            op = int("".join(n)), "".join(next(cig_iter)[1])
            if op[1] in self._valid_cigar:
                yield op
            else:
                raise ValueError("CIGAR operation %s in record %s is invalid." % (op[1], self.qname))

    def gapped(self, attr, gap_char='-'):
        """ Return a :class:`.Sam` sequence attribute or tag with all
        deletions in the reference sequence represented as 'gap_char' and all
        insertions in the reference sequence removed. A sequence could
        be :class:``Sam.seq``, ``Sam.qual``, or any :class:`.Sam` tag that
        represents an aligned sequence, such as a methylation tag for bisulfite
        sequencing libraries.

        >>> x = Sam(*'r001\t99\tref\t7\t30\t8M2I4M1D3M\t=\t37\t39\tTTAGATAAAGGATACTG\t*'.split())
        >>> x.gapped('seq')
        'TTAGATAAGATA-CTG'
        >>> x = Sam(*'r001\t99\tref\t7\t30\t8M2I4M1D3M\t=\t37\t39\tTTAGATAAAGGATACTG\t*'.split(), tags=['ZM:Z:.........M....M.M'])
        >>> x.gapped('ZM')
        '............-M.M'
        """
        try:
            ungapped = getattr(self, attr)
        except AttributeError:
            ungapped = self[attr]  # get dictionary key (tag) if attribute is missing
        if len(ungapped) != len(self.seq):
            raise ValueError("The length of the '%s' attribute is not equal to the length of Sam.seq!" % attr) 
        gapped = []
        i = 0
        for n, t in self.cigars:
            if t in self._cigar_align:
                gapped.extend(ungapped[i:i + n])
                i += n
            elif t in self._cigar_ref_only:
                gapped.extend([gap_char] * n)
            elif t in self._cigar_query_only:
                i += n
            elif t in self._cigar_no_align:
                pass
        return ''.join(gapped)

    def parse_md(self):
        """ Return the ungapped reference sequence from the MD tag, if present.
        """
        try:
            return self._cache['parse_md']
        except KeyError:
            pass
        try:
            md = self['MD']
        except KeyError:
            raise KeyError('MD tag not found in SAM record.')
        ref_seq = list(self.gapped('seq'))
        md_match = re.findall(r"([0-9]+)\^?([A-Z]+)?", md)
        ref_seq_i = 0
        for i, b in md_match:
            ref_seq_i += int(i)
            for mismatch in b:
                try:
                    ref_seq[ref_seq_i] = mismatch
                except IndexError:
                    raise IndexError(locals())
                ref_seq_i += 1
        self._cache['parse_md'] = ref_seq
        return ref_seq

    @property
    def cigars(self):
        """ Returns the CIGAR string as a tuple.

        >>> x = Sam(cigar='8M2I4M1D3M')
        >>> x.cigars
        ((8, 'M'), (2, 'I'), (4, 'M'), (1, 'D'), (3, 'M'))
        """
        try:
            return self._cache['cigars']
        except KeyError:
            self._cache['cigars'] = tuple(self.cigar_split())
            return self._cache['cigars']

    @property
    def tags(self):
        """ Parses the tags string to a dictionary if necessary.

        >>> x = Sam(tags=['XU:Z:cgttttaa', 'XB:Z:cttacgttaagagttaac', 'MD:Z:75', 'NM:i:0', 'NH:i:1', 'RG:Z:1'])
        >>> sorted(x.tags.items(), key=lambda x: x[0])
        [('MD', '75'), ('NH', 1), ('NM', 0), ('RG', '1'), ('XB', 'cttacgttaagagttaac'), ('XU', 'cgttttaa')]
        """
        try:
            return self._cache['tags']
        except KeyError:
            self._cache['tags'] = parse_sam_tags(self._tags)
            return self._cache['tags']

    @property
    def paired(self):
        """ Returns True if the read is paired and
        each segment properly aligned according to the aligner. """
        return bool(self.flag & 0x2)

    @property
    def mapped(self):
        """ Returns True of the read is mapped. """
        return not (self.flag & 0x4)

    @property
    def secondary(self):
        """ Returns True if the read alignment is secondary. """
        return bool(self.flag & 0x100)

    @property
    def reverse(self):
        """ Returns True if ``Sam.seq`` is being reverse complemented. """
        return bool(self.flag & 0x10)

    @property
    def passing(self):
        """ Returns True if the read is passing filters, such as platform/vendor quality controls. """
        return not bool(self.flag & 0x200)

    @property
    def duplicate(self):
        """ Returns True if the read is a PCR or optical duplicate. """
        return bool(self.flag & 0x400)

    @property
    def coords(self):
        """ Return a list of genomic coordinates for the gapped alignment. """
        return range(self.pos, self.pos + len(self))

    @property
    def safename(self):
        """Return ``Sam.qname`` without paired-end identifier if it exists"""
        if self.qname[-2] == '/':
            return self.qname[:-2]
        else:
            return self.qname


def parse_sam_tags(tagfields):
    """ Return a dictionary containing the tags """
    return dict([(tag, data) for tag, dtype, data in [decode_tag(x) for x in tagfields]])


def encode_tag(tag, data):
    """ Write a SAM tag in the format ``TAG:TYPE:data``. Infers the data type
    from the Python object type.

    >>> encode_tag('YM', '#""9O"1@!J')
    'YM:Z:#""9O"1@!J'
    """
    if isinstance(data, string_types):
        data_type = 'Z'
    elif isinstance(data, int):
        data_type = 'i'
    elif isinstance(data, float):
        data_type = 'f'
    else:
        raise NotImplementedError("Data {0} cannot be encoded as string, integer, or float tag.".format(data))
    value = ':'.join((tag, data_type, str(data)))
    return value


def decode_tag(tag_string):
    """ Parse a SAM format tag to a (tag, type, data) tuple. Python object
    types for data are set using the type code. Supported type codes are: A, i, f, Z, H, B

    >>> decode_tag('YM:Z:#""9O"1@!J')
    ('YM', 'Z', '#""9O"1@!J')
    >>> decode_tag('XS:i:5')
    ('XS', 'i', 5)
    >>> decode_tag('XF:f:100.5')
    ('XF', 'f', 100.5)
    """
    try:
        tag, data_type, data = tag_string.split(':')
    except ValueError:
        match = re.match(r'([A-Z]{2}):([iZfHB]):(\S+)', tag_string)
        tag = match.group(1)
        data_type = match.group(2)
        data = match.group(3)
    if data_type == 'i':
        return (tag, data_type, int(data))
    elif data_type == 'Z':
        return (tag, data_type, data)
    elif data_type == 'f':
        return (tag, data_type, float(data))
    elif data_type == 'A':  # this is just a special case of a character
        return (tag, data_type, data)
    elif data_type == 'H':
        raise NotImplementedError("Hex array SAM tags are currently not parsed.")
    elif data_type == 'B':
        raise NotImplementedError("Byte array SAM tags are currently not parsed.")
    else:
        raise NotImplementedError("Tag {0} cannot be parsed.".format(tag_string))


def tile_region(rname, start, end, step):
    """ Make non-overlapping tiled windows from the specified region in
    the UCSC-style string format.

    >>> list(tile_region('chr1', 1, 250, 100))
    ['chr1:1-100', 'chr1:101-200', 'chr1:201-250']
    >>> list(tile_region('chr1', 1, 200, 100))
    ['chr1:1-100', 'chr1:101-200']
    """
    while start + step <= end:
        yield '%s:%d-%d' % (rname, start, start + step - 1)
        start += step
    if start < end:
        yield '%s:%d-%d' % (rname, start, end)

def bam_read_count(bamfile, samtools_path="samtools"):
    """ Return a tuple of the number of mapped and unmapped reads in a BAM file """
    p = Popen([samtools_path, 'idxstats', bamfile], stdout=PIPE)
    mapped = 0
    unmapped = 0
    for line in p.stdout:
        rname, rlen, nm, nu = line.rstrip().split()
        mapped += int(nm)
        unmapped += int(nu)
    return (mapped, unmapped)


class fragment(object):
    def __init__(self, loc, bed_start, end, fraglen, R1_mapq, R2_mapq, R1_flag, R2_flag, direction):
        self.loc = loc
        self.bed_start = bed_start
        self.end = end
        self.fraglen = fraglen
        self.R1_mapq = R1_mapq
        self.R2_mapq = R2_mapq
        self.R1_flag = R1_flag
        self.R2_flag = R2_flag
        self.direction = direction
    def __repr__(self):
        return "%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n" % (self.loc,self.bed_start,self.end,self.fraglen,self.R1_mapq,self.R2_mapq,self.R1_flag, self.R2_flag, self.direction)
    def __str__(self):
        return "%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n" % (self.loc,self.bed_start,self.end,self.fraglen,self.R1_mapq,self.R2_mapq,self.R1_flag, self.R2_flag, self.direction)

class fragments(object):
    def __init__(self, in_file, out_file=None, skip_dups = True, filter_threshes = None):
        #self.in_file = in_file
        self.read_count=0
        self.fragment_count = 0
        if filter_threshes is None:
            self.filter_threshes = [float("-inf"),float("inf")]
        else:
            self.filter_threshes = filter_threshes
        self.passed_filter = 0
        if skip_dups:
            self.forward_list = [99, 163, 355, 419]
            self.plus_list = [99, 355]
        else:
            self.forward_list = [99, 163, 355, 419, 1123, 1187]
            self.plus_list = [99, 355, 1123]
        if out_file is None:
            self.return_fragments(in_file)
        else:
            self.write_fragments(in_file, out_file)
        # self.generator = self.parse_sam_file(self.in_sam)

    def load_fragments(self, in_file):

        in_sam = Reader(in_file)
        iterator = iter(in_sam)
        done_looping = False
        fragments = []
        while not done_looping:
            try:
                read1 = next(iterator)
                self.read_count += 1
                read2 = next(iterator)
                self.read_count += 1
            except StopIteration:
                done_looping = True
                #print("Processed %s reads.  Found %s usable fragments" % (self.read_count, self.fragment_count))
            else:
                return_val = self.qualify_reads(read1, read2)
                if return_val is not None:
                    fragments.append(return_val)
                #print(str(self.qualify_reads(read1, read2)))
        return fragments

    def return_fragments(self, in_file):
        in_sam = Reader(in_file)
        iterator = iter(in_sam)
        done_looping = False
        while not done_looping:
            try:
                read1 = next(iterator)
                self.read_count += 1
                read2 = next(iterator)
                self.read_count += 1
            except StopIteration:
                done_looping = True
            except IOError:
                done_looping = True
                #print("Processed %s reads.  Found %s usable fragments" % (self.read_count, self.fragment_count))
            else:
                return_val = self.qualify_reads(read1, read2)
                if return_val is not None:
                    sys.stdout.write(str(return_val))
                #print(str(self.qualify_reads(read1, read2)))

    def write_fragments(self, in_file, out_file):

        in_sam = Reader(in_file)
        iterator = iter(in_sam)
        done_looping = False
        fragments = []
        while not done_looping:
            try:
                read1 = next(iterator)
                self.read_count += 1
                read2 = next(iterator)
                self.read_count += 1
            except StopIteration:
                done_looping = True
                print("\n[SAM2BED] Output: \n Processed %s read pairs.  Found %s usable fragments.\nOf these, %s passed_filter.\n" % (self.read_count/2, self.fragment_count, self.passed_filter))
            else:
                return_val = self.qualify_reads(read1, read2, filter = self.filter_threshes)
                if return_val is not None:
                    out_file.writelines(str(line) for line in [return_val])

    # def qualify_reads_old(self, read1, read2, filter = [float("-inf"),float("inf")]):
    #     direction = "Unk"
    #     r1_unique = r2_unique = False
    #     if read1.flag in self.forward_list:
    #         if "XS" not in read1.tags:
    #             r1_unique=True
    #             if "XS" not in read2.tags:
    #                 r2_unique=True
    #                 begin = read1.pos
    #                 end = (read2.pos + len(read2.seq))
    #                 direction = "+"
    #     if read2.flag in self.forward_list:
    #         if "XS" not in read2.tags:
    #             r2_unique=True
    #             if "XS" not in read1.tags:
    #                 r1_unique=True
    #                 begin = read2.pos
    #                 end = read1.pos - len(read1.seq)
    #                 direction = "-"
    #     if r1_unique and r2_unique and direction is not "Unk" and (end - begin) >0 :
    #         self.fragment_count+=1
    #         if (end - begin) > filter[0] and (end - begin) < filter[1]:
    #             self.passed_filter+=1
    #             return fragment(read1.rname, begin, end, end - begin, read1.mapq, read2.mapq, read1.flag, read2.flag, direction)

    # Sam fields: qname='', flag=4, rname='*', pos=0, mapq=255, cigar='*', rnext='*', pnext=0, tlen=0, seq='*', qual='*', tags=[]
    def qualify_reads(self, read1, read2, filter = [float("-inf"),float("inf")]):
        direction = "Unk"
        r1_unique = r2_unique = False
        if read1.flag in self.forward_list:
            if "XS" not in read1.tags:
                r1_unique=True
                if "XS" not in read2.tags:
                    r2_unique=True
                    begin = read1.pos
                    end = read1.pos + read1.tlen
                    direction = "+"
        if read2.flag in self.forward_list:
            if "XS" not in read2.tags:
                r2_unique=True
                if "XS" not in read1.tags:
                    r1_unique=True
                    begin = read2.pos
                    end = read2.pos + read2.tlen
                    direction = "-"
        if r1_unique and r2_unique and direction is not "Unk" and (end - begin) >0 and read1.rnext == "=" and read2.rnext == "=":
            self.fragment_count+=1
            if (end - begin) > filter[0] and (end - begin) < filter[1]:
                self.passed_filter+=1
                return fragment(read1.rname, (begin -1) , (end - 1) , end - begin, read1.mapq, read2.mapq, read1.flag, read2.flag, direction)


def run_sam2bed():
    parser = argparse.ArgumentParser('A script for converting sam file to bed for eventual bedgraph conversion')
    parser.add_argument('sam_bam', type=str, default = sys.stdin, help='sam or bam fil or stdin')
    parser.add_argument('--out_bed', '-o', default = None, help='bed output', required=False)
    parser.add_argument('--filter_high', '-fh', type = int, default = None, help='filter_high', required=False)
    parser.add_argument('--filter_low', '-fl', default = None, type = int,  help='filter_high', required=False)
    args = parser.parse_args()
    if args.sam_bam=="-":
        args.sam_bam = sys.stdin
    else:
        args.sam_bam = open(args.sam_bam, 'r')

    filter_threshes = []
    if args.filter_low is not None:
        filter_threshes.append(args.filter_low)
    else:
        filter_threshes.append(float("-inf"))
    if args.filter_high is not None:
        filter_threshes.append(args.filter_high)
    else:
        filter_threshes.append(float("inf"))


    if args.out_bed is not None:
        of = open(args.out_bed, "w")
        fragments(args.sam_bam, out_file=of, filter_threshes = filter_threshes)
        of.close()
    else:
        fragments(args.sam_bam, filter_threshes = filter_threshes)

if __name__ == '__main__':
    run_sam2bed()







#!/usr/bin/python
#v0.2

import sys
import argparse


# inFileA = args.sam1
# inFileB = args.sam2

def samComp_writer(inFileA, inFileB, outfile, mode):
    reads = {}
    for line in inFileA:
        #deal with header
        if line[0] != "@": 
            data = line.strip().split('\t')
            name = data[0]
            flag = int(data[1])
            if not flag & 4:
                reads[name] = True

    if mode=="intersect":
        for line in inFileB:
            #deal with header
            if line[0] != "@": 
                data = line.strip().split('\t')
                name = data[0]
                flag = int(data[1])
                if not flag & 4:
                    if name in reads:
                        outfile.writelines('\t'.join(data)+'\n')

    if mode=="unique":
        for line in inFileB:
            #deal with header
            if line[0] != "@":
                data = line.strip().split('\t')
                name = data[0]
                flag = int(data[1])
                if not flag & 4:
                    if name not in reads:
                        outfile.writelines('\t'.join(data)+'\n')
    return


def samComp():
    parser = argparse.ArgumentParser('A script finding either the "intersect" or "unique" reads in two SAM files (for "unique": i.e reads in sam2 not in sam1')
    parser.add_argument('sam1', type=str, default = sys.stdin, help='reference sam')
    parser.add_argument('sam2', type=str, default = sys.stdin, help='reference sam')
    parser.add_argument('--out_sam', '-o', default = "./out.sam", help='SAM output', required=False)
    parser.add_argument('--mode', '-m', type = str, default = "unique", choices=['unique', 'intersect'], help='mode to compare sam files', required=False)
    args = parser.parse_args()
    """
    call='samComp /Users/sfurla/Desktop/smaller.sam /Users/sfurla/Desktop/smaller.sam -m intersect'
    args = parser.parse_args(call.split(" ")[1:])
    """

    #Deal with pipes and open files
    if args.sam1=="-" and args.sam2=="-":
        raise ValueError("Two pipes not supported")
    if args.sam1=="-":
        args.sam1 = sys.stdin
    else:
        args.sam1 = open(args.sam1, 'r')

    if args.sam2=="-":
        args.sam2 = sys.stdin
    else:
        args.sam2 = open(args.sam2, 'r')

    #open write file
    args.out_sam = open(args.out_sam, 'w')

    samComp_writer(args.sam1, args.sam2, args.out_sam, args.mode)

if __name__ == '__main__':
    samComp()







import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

    setuptools.setup(
    name="henipipe",
    version="1.64",
    author="Scott Furlan",
    author_email="scottfurlan@gmail.com",
    description="A python wrapper for fast and parallel processing of sequencing data using CutnRun or CutnTag",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/scfurl/henipipe.git",
    packages=setuptools.find_packages(),
    package_data={'henipipe': ['henipipe/data/genomes.json']},
    install_requires=['six >= 1'],
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=2.5',
    entry_points={'console_scripts': [
        'henipipe = henipipe.__main__:run_henipipe',
        'samTobed = henipipe.samTobed:run_sam2bed',
        'pyWriter = henipipe.pyWriter:run_pyWriter',
        'auc = henipipe.auc:run_auc',
    ]},
    )


"""
#TEST PyPI
## run this to make package: python3 setup.py sdist bdist_wheel
## run this to upload to test pypi: twine upload --repository-url https://test.pypi.org/legacy/ dist/*
#TEST INSTALL
# pip install --index-url https://test.pypi.org/henipipe/ henipipe


#FOR REAL
cd ~/OneDrive\ -\ Fred\ Hutchinson\ Cancer\ Research\ Center/computation/develop/henipipe/
rm dist/*
python3 setup.py sdist bdist_wheel
python3 -m twine upload dist/*
scfurl


##Install pipx
## python3 -m pip install --user pipx
## python3 -m pipx ensurepath

**At SCRI do the following**

module load python
python3 -m pip install --user pipx
python3 -m pipx ensurepath
pipx install --include-deps --pip-args '--trusted-host pypi.org --trusted-host files.pythonhosted.org' henipipe
pipx install --spec git+https://github.com/scfurl/henipipe --include-deps henipipe --pip-args '--trusted-host pypi.org --trusted-host files.pythonhosted.org'
pipx install --spec git+https://github.com/scfurl/henipipe@cleaner --include-deps henipipe --pip-args '--trusted-host pypi.org --trusted-host files.pythonhosted.org'


**At the FHCRC do the following...**

module load Python/3.6.7-foss-2016b-fh1
python3 -m pip install --user pipx
python3 -m pipx ensurepath
pipx install --include-deps henipipe
pipx install --spec git+https://github.com/scfurl/henipipe --include-deps henipipe 
pipx uninstall henipipe


pipx uninstall henipipe
pipx install git+https://github.com/scfurl/henipipe --include-deps

"""


#!/usr/bin/env python

#AUC finder
#this script will find AUC measurements using ranges taken from a peak_file and interrogating a number of target files
#requires tabix

#whole_peak and narrow_peak
import argparse
import os
from subprocess import Popen, PIPE
import time




class AUC:
    def __init__(self, *args, **kwargs):
        self.peak_file = kwargs.get('peak_file')
        self.targets = kwargs.get('targets')
        self.file_out = kwargs.get('file_out')
        self.header = kwargs.get('header')
        self.peak_count=0

    def __call__():
        pass

    def address_ok(self, address):
        first_split = address.split(":")
        locs = first_split[1].split("-")
        if any([not s for s in first_split]) or any([not s for s in locs]):
            return False
        else:
            return True

    def peak_line(self, peak):
        address = peak.split("\t")[:3]
        wide_peak = "{0:s}:{1:s}-{2:s}".format(*address)
        narrow_peak = peak.split("\t")[5].split()[0]
        #i = targets[0]
        #print(wide_peak)
        #print(narrow_peak)
        if self.address_ok(wide_peak) and self.address_ok(narrow_peak):
            wide_values=[]
            narrow_values=[]
            for i in self.targets:
                proc = Popen("tabix {0:s} {1:s} | awk '{{s+=$4}} END {{print s}}'".format(i, wide_peak), shell = True, stdin = PIPE, stdout=PIPE, stderr = PIPE)
                out, err = proc.communicate()
                try:
                    wide_values.append(out.decode('UTF-8').split()[0])
                except IndexError:
                    wide_values.append("NF")
                proc = Popen("tabix {0:s} {1:s} | awk '{{s+=$4}} END {{print s}}'".format(i, narrow_peak), shell = True, stdin = PIPE, stdout=PIPE, stderr = PIPE)
                out, err = proc.communicate()
                try:
                    narrow_values.append(out.decode('UTF-8').split()[0])
                except IndexError:
                    narrow_values.append("NF")
            return "\t".join(["\t".join(address), ("\t".join(wide_values)), ("\t".join(narrow_values)), narrow_peak])+"\n"
        if self.address_ok(wide_peak) and not self.address_ok(narrow_peak):
            wide_values=[]
            print("Narrow peak error found on line {0} of peak file".format(self.peak_count))
            for i in self.targets:
                proc = Popen("tabix {0:s} {1:s} | awk '{{s+=$4}} END {{print s}}'".format(i, wide_peak), shell = True, stdin = PIPE, stdout=PIPE, stderr = PIPE)
                out, err = proc.communicate()
                try:
                    wide_values.append(out.decode('UTF-8').split()[0])
                except IndexError:
                    wide_values.append("NF")
            return "\t".join(["\t".join(address), ("\t".join(wide_values)), ("\t".join(["NF"]*len(self.targets))), narrow_peak])+"\n"
        if not self.address_ok(wide_peak):
            print("Wide peak error found on line {0} of peak file".format(self.peak_count))
            return None

    def calculate_AUCs(self, start=1, end=1e6):
        """
        peak_file = open(args.peak_file, 'r')
        targets = args.targets
        file_out = open(args.output, 'w')
        """
        start_time = time.time()
        peak_file = open(self.peak_file, 'r')
        file_out = open(self.file_out, 'w')
        if self.header:
            target_bn_wide = "\t".join([os.path.basename(i)+"_wide" for i in self.targets])
            target_bn_narrow = "\t".join([os.path.basename(i)+"_narrow" for i in self.targets])
            header = "seq\tbegin\tend\t"+target_bn_wide+"\t"+target_bn_narrow+"\tnarrow_coord\n"
            file_out.writelines(header)
        iterator = iter(peak_file)
        done_looping = False
        while not done_looping:
            try:
                peak = next(iterator)
                self.peak_count += 1
                #print(self.peak_count)
                if self.peak_count >= end:
                    done_looping = True #2265 error
                    print("\n[AUC] Output: \n Processed {0} peaks\n".format(self.peak_count))
            except StopIteration:
                peak_file.close()
                file_out.close()
                print("\n[AUC] Output: \n Processed {0} peaks".format(self.peak_count))
                done_looping = True
                end_time = time.time()
                print("\n[AUC] Output: \n Execution took {0} seconds\n".format(end_time - start_time))
            else:
                #line_out = peak_line(peak, args.targets)
                if self.peak_count >= start:
                    line_out = self.peak_line(peak)
                    if line_out is not None: file_out.writelines(line_out)
        return


def run_auc():
    parser = argparse.ArgumentParser('A script that will find AUC measurements using ranges taken from a peak_file and interrogating a number of target files')
    parser.add_argument('--peak_file', '-p', type=str, help='A peak file (only SEACR supported)')
    parser.add_argument('--output', '-o', type=str, default=".", help='Write files to this location')
    parser.add_argument('--keepfiles', '-k', action ='store_true', default=False,  help='Write files to this location')
    parser.add_argument('--noheader', '-nh', action ='store_true', default=False,  help='Do not include a header')
    parser.add_argument('targets', nargs='*')
    args = parser.parse_args()
    """
    call = 'auc -o fc/4N1_H3K27me3_4G1_H3K27me3_AUC.bed -p /active/furlan_s/Data/CNR/190801_CNRNotch/henipipe_150/fc/4N1_H3K27me3_4G1_H3K27me3_SEACR.stringent.bed /active/furlan_s/Data/CNR/190801_CNRNotch/henipipe_150/fc/4N1_H3K27me3.bedgraph.bz /active/furlan_s/Data/CNR/190801_CNRNotch/henipipe_150/fc/4G1_H3K27me3.bedgraph.bz'
    args = parser.parse_args(call.split(" ")[1:])
    """

    if os.path.isabs(args.output) is False:
        args.output = os.path.abspath(args.output)

    AUCjob = AUC(peak_file = args.peak_file, targets = args.targets, file_out = args.output, header = not args.noheader)
    # AUCjob.targets
    # AUCjob.peak_file
    AUCjob = AUCjob.calculate_AUCs()

if __name__ == '__main__':
    run_auc()




